# Test2.py

from gval import CatStats as CatSt
# from gval.statistics.test import cat_stats, register


if __name__ == "__main__":
    # Method 1 - Import from another file
    print(CatSt.available_functions())
    print(CatSt.matthews_correlation_coefficient(100, 0, 0, 0))
    # print(cat_stats.available_functions())
    # print("test", cat_stats.registered_functions["test"])
    # print(cat_stats.test(*[30, 20]))
    # print("test6", cat_stats.registered_functions["test6"])
    # print(cat_stats.test6(*[30, 20]))
    #
    # # Method 2 - Call register function from separate file
    # new_obj = CatSt()
    # register(new_obj)
    # print(new_obj.available_functions())
    # print("test", new_obj.registered_functions["test"])
    # print(new_obj.test(*[30, 20]))
    # print("test6", new_obj.registered_functions["test6"])
    # print(new_obj.test6(*[30, 20]))


# pending registration discusison
# persistence (disk? config? cache?)
# pending comments xarray accessor issue discussion
# pending jupyter notebook



import gval
from gval.comparison.pairing_functions import _make_pairing_dict
import rioxarray as rxr
import xarray as xr
import numpy as np
import geopandas as gpd
from geocube.api.core import make_geocube


def run():
    c = rxr.open_rasterio('/home/sven/repos/fim/inputs/test_cases/ras2fim_test_cases/07100006_ras2fim/testing_versions/test4/5yr/inundation_extent_07100006.tif', mask_and_scale=True)
    b = rxr.open_rasterio('/home/sven/repos/fim/inputs/test_cases/ras2fim_test_cases/validation_data_ras2fim/07100006/5yr/ras2fim_huc_07100006_extent_5yr.tif', mask_and_scale=True)

    c.data = xr.where(c >= 0, 1, c)
    c.data = xr.where(c < 0, 0, c)

    mask_dict = {'levees': {'path': '~/repos/fim/inputs/inputs/nld_vectors/Levee_protected_areas.gpkg',
                            'buffer': None, 'operation': 'exclude'},
                 'waterbodies': {'path': '~/repos/fim/inputs/inputs/nwm_hydrofabric/nwm_lakes.gpkg',
                                 'buffer': None, 'operation': 'exclude'}}

    df = gpd.read_file(mask_dict['levees']['path'], bbox=c.rio.bounds())
    df['mask'] = 4
    df2 = gpd.read_file(mask_dict['waterbodies']['path'], bbox=c.rio.bounds())
    df2['mask'] = 4

    e = make_geocube(df, ['mask'], like=c)
    f = make_geocube(df2, ['mask'], like=c)
    g = xr.merge([e, f]).to_array()

    c.data = xr.where(g.data == 4, 4, c)

    pairing_dictionary = {(0.0, 0.0): 0,
     (0.0, 1.0): 1,
     (0.0, np.nan): np.nan,
     (1.0, 0.0): 2,
     (1.0, 1.0): 3,
     (1.0, np.nan): np.nan,
     (4.0, 0.0): 4,
     (4.0, 1.0): 4,
     (4.0, np.nan): np.nan,
     (np.nan, 0.0): np.nan,
     (np.nan, 1.0): np.nan,
     (np.nan, np.nan): np.nan}


    aaa, bbb, ccc = c.gval.categorical_compare(b, positive_categories=[1],
                                          negative_categories=[0],
                                          comparison_function='pairing_dict',
                                          allow_candidate_values=[0, 1, 4],
                                          allow_benchmark_values=[0, 1, 4],
                                          pairing_dict=pairing_dictionary)

if __name__ == '__main__':

    run()


"""
Computes continuous value metrics given an agreement map.
"""

# __all__ = ['*']
__author__ = "Fernando Aristizabal"

from typing import Iterable, Union

import pandera as pa
import pandas as pd
from pandera.typing import DataFrame
import xarray as xr
import geopandas as gpd

from gval import ContStats
from gval.utils.schemas import Metrics_df, Subsample_identifiers, Sample_identifiers


@pa.check_types
def _compute_continuous_metrics(
    agreement_map: Union[xr.DataArray, xr.Dataset],
    candidate_map: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
    metrics: Union[str, Iterable[str]] = "all",
    subsampling_average: str = "micro",
    subsampling_df: gpd.GeoDataFrame = None,
) -> DataFrame[Metrics_df]:
    """
    Computes categorical metrics from a crosstab df.

    Parameters
    ----------
    agreement_map : Union[xr.DataArray, xr.Dataset, List[Union[xr.DataArray, xr.Dataset]]]
        Agreement map, error based (candidate - benchmark).
    candidate_map : Union[xr.DataArray, xr.Dataset]
        Candidate map.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map.
    metrics : Union[str, Iterable[str]], default = "all"
        String or list of strings representing metrics to compute.
    subsampling_average : str, default = "micro"
        Strategy to average samples if there is more than one in the agreement map
    subsampling_df : gpd.GeoDataFrame, default = None
        DataFrame with geometries to subsample or use as exclusionary masks and optional sample weights

    Returns
    -------
    DataFrame[Metrics_df]
        Metrics DF with computed metrics per sample.

    Raises
    ------
    ValueError
        If metrics is not a string or list of strings.

    References
    ----------
    .. [1] `7th International Verification Methods Workshop <https://www.cawcr.gov.au/projects/verification/>`_
    .. [2] `3.3. Metrics and scoring: quantifying the quality of predictions <https://scikit-learn.org/stable/modules/model_evaluation.html>`_
    """

    if not isinstance(agreement_map, list):
        agreement_map = [agreement_map]
        candidate_map = [candidate_map]
        benchmark_map = [benchmark_map]

    metric_dfs = []
    for idx, (agreement, benchmark, candidate) in enumerate(
        zip(agreement_map, benchmark_map, candidate_map)
    ):
        # compute error based metrics such as MAE, MSE, RMSE, etc. from agreement map and produce metrics_df
        statistics, names = ContStats.process_statistics(
            metrics,
            error=agreement,
            candidate_map=candidate,
            benchmark_map=benchmark,
        )

        # create metrics_df
        metric_df = dict()
        for name, stat in zip(names, statistics):
            metric_df[name] = stat

        def is_nested_dict(d):
            # if not isinstance(d, dict):
            #     return False
            return any(isinstance(v, dict) for v in d.values())

        if is_nested_dict(metric_df):
            metric_df = pd.DataFrame.from_dict(metric_df, orient="index").transpose()
            metric_df.reset_index(inplace=True)
            metric_df.rename(columns={"index": "band"}, inplace=True)
            metric_df["band"] = metric_df["band"].str.replace("band_", "")

        else:
            # dataarray
            metric_df = pd.DataFrame(metric_df, index=[0])

            # add band
            metric_df.insert(0, "band", "1")

        if subsampling_df is not None:
            metric_df.insert(0, "subsample", f"{idx + 1}")

        metric_dfs.append(metric_df)

    metric_df = pd.concat(metric_dfs).reset_index().drop(columns=["index"])

    if subsampling_df is not None:
        if subsampling_average == "band":
            metric_df = (
                metric_df.groupby(Subsample_identifiers.columns())
                .mean(numeric_only=True)
                .reset_index()
            )

            metric_df.insert(1, "band", "averaged")

        if subsampling_average == "subsample":
            metric_df = (
                metric_df.groupby(Sample_identifiers.columns())
                .mean(numeric_only=True)
                .reset_index()
            )

            metric_df.insert(0, "subsample", "averaged")

        if subsampling_average == "weighted":
            if subsampling_df.get("weights") is None:
                raise ValueError(
                    "Must have weights if weighted is chosen for subsampling"
                )  # pragma: no cover

            metric_df.loc[:, "weights"] = subsampling_df["weights"]

            # compute weighted average
            weighted_metrics = (
                metric_df.loc[:, metrics]
                .multiply(metric_df.loc[:, "weights"], axis=0)
                .reset_index(drop=True)
            )

            # add weighted metrics to metric_df
            metric_df.loc[:, metrics] = weighted_metrics

            # take average of weighted metrics
            metric_df = (
                metric_df.groupby(Sample_identifiers.columns())
                .sum(numeric_only=True)
                .drop(
                    columns=["weights", "subsample"],
                    errors="ignore",
                )
                .divide(metric_df.loc[:, "weights"].sum())
                .reset_index()
            )

            metric_df.insert(0, "subsample", "averaged")

    return metric_df



import memray
import gval
import rioxarray as rxr
import subprocess
import os


def run():
    candidate = rxr.open_rasterio('candidate_map_two_class_categorical.tif', mask_and_scale=True)
    benchmark = rxr.open_rasterio('benchmark_map_two_class_categorical.tif', mask_and_scale=True)

    candidate.gval.categorical_compare(benchmark,
                                                 positive_categories=[2],
                                                 negative_categories=[0, 1])


if __name__ == '__main__':

    file_name = "output.bin"

    if os.path.exists(file_name):
        os.remove(file_name)

    with memray.Tracker(file_name):
        run()

    subprocess.call(['memray', 'summary', file_name])




from gval.utils.loading_datasets import handle_xarray_memory, adjust_memory_strategy
import rioxarray as rxr
from gval.statistics.test4 import a

if __name__ == '__main__':

    adjust_memory_strategy('aggressive')
    multi_cand = rxr.open_rasterio('s3://gval-test/candidate_categorical_multiband_aligned_0.tif', mask_and_scale=True,
                                   band_as_variable=True)

    multi_cand = handle_xarray_memory(multi_cand)
    a(multi_cand)



"""
Functions to load datasets
"""

# __all__ = ['*']
__author__ = "Fernando Aristizabal"

from typing import Union, Optional, Tuple, Dict, Any
import os
import ast

import rioxarray as rxr
import xarray as xr
import rasterio
from tempfile import NamedTemporaryFile
from rio_cogeo.cogeo import cog_translate
from rio_cogeo.profiles import cog_profiles


def load_raster_as_xarray(
    filename: Union[
        str, os.PathLike, rasterio.io.DatasetReader, rasterio.vrt.WarpedVRT
    ],
    parse_coordinates: Optional[bool] = None,
    chunks: Optional[Union[int, Tuple, Dict]] = None,
    cache: Optional[bool] = None,
    lock: Optional[Any] = None,
    masked: Optional[bool] = False,
    mask_and_scale: Optional[bool] = False,
    default_name: Optional[str] = None,
    band_as_variable: Optional[bool] = False,
    **open_kwargs,
) -> Union[xr.DataArray, xr.Dataset]:
    """
    Wraps around :obj:`rioxarray.open_rasterio` providing control over some arguments.

    .. deprecated:: 0.0.2
        `load_raster_as_xarray` will be removed in gval 0.0.2.  Use `rioxarray.open_rasterio` instead

    Parameters
    ----------
    filename : Union[ str, os.PathLike, rasterio.io.DatasetReader, rasterio.vrt.WarpedVRT ]
        Path to the file to open. Or already open rasterio dataset
    parse_coordinates : Optional[bool], default = None
        Whether to parse the x and y coordinates out of the file's
        ``transform`` attribute or not. The default is to automatically
        parse the coordinates only if they are rectilinear (1D).
        It can be useful to set ``parse_coordinates=False``
        if your files are very large or if you don't need the coordinates.
    chunks : Optional[Union[int, Tuple, Dict]], default = None
        Chunk sizes along each dimension, e.g., ``5``, ``(5, 5)`` or
        ``{'x': 5, 'y': 5}``. If chunks is provided, it used to load the new
        DataArray into a dask array. Chunks can also be set to
        ``True`` or ``"auto"`` to choose sensible chunk sizes according to
        ``dask.config.get("array.chunk-size")``.
    cache : Optional[bool], default = None
        If True, cache data loaded from the underlying datastore in memory as
        NumPy arrays when accessed to avoid reading from the underlying data-
        store multiple times. Defaults to True unless you specify the `chunks`
        argument to use dask, in which case it defaults to False.
    lock : Optional[Any], default = None
        If chunks is provided, this argument is used to ensure that only one
        thread per process is reading from a rasterio file object at a time.

        By default and when a lock instance is provided,
        a :class:`xarray.backends.CachingFileManager` is used to cache File objects.
        Since rasterio also caches some data, this will make repeated reads from the
        same object fast.

        When ``lock=False``, no lock is used, allowing for completely parallel reads
        from multiple threads or processes. However, a new file handle is opened on
        each request.
    masked: Optional[bool], default = False
        If True, read the mask and set values to NaN. Defaults to False.
    mask_and_scale: Optional[bool], default = False
        Lazily scale (using the `scales` and `offsets` from rasterio) and mask.
        If the _Unsigned attribute is present treat integer arrays as unsigned.
    default_name : Optional[str], default = None
        The name of the data array if none exists. Default is None.
    band_as_variable : Optional[bool], default = False
        If True, will load bands in a raster to separate variables.
    open_kwargs : kwargs, default = None
        Optional keyword arguments to pass into :func:`rasterio.open`.


    Returns
    -------
    Union[:obj:`xr.DataArray`, :obj:`xr.Dataset`]
        Loaded data.

    References
    ----------
    .. [1] `rioxarray.open_rasterio() <https://corteva.github.io/rioxarray/stable/rioxarray.html>`_
    .. [2] `rasterio.open() <https://rasterio.readthedocs.io/en/stable/api/rasterio.html#rasterio.open>`_
    """

    return rxr.open_rasterio(
        filename=filename,
        parse_coordinates=parse_coordinates,
        cache=cache,
        lock=lock,
        default_name=default_name,
        band_as_variable=band_as_variable,
        masked=masked,
        chunks=chunks,
        mask_and_scale=mask_and_scale,
        **open_kwargs,
    )


_MEMORY_STRATEGY = "normal"


def adjust_memory_strategy(strategy: str):
    """
    Tells GVAL how to address handling memory.  There are three modes currently available:

    normal:  Keeps all of xarray files in memory as usual
    moderate:  Either creates cloud optimized geotiffs and stores as temporary files and reloads or reloads file
    to be in lazily loaded stated
    aggressive: Does the same as moderate except loads with no cache so everything is read from disk

    There are tradeoffs with performance for choosing a strategy that conserves memory, adjust only as needed.

    Parameters
    ----------
    strategy : str, {'normal', 'moderate', 'aggressive'}
        Method to conserve memory

    Raises
    ------
    ValueError

    """

    if strategy in {"normal", "moderate", "aggressive"}:
        global _MEMORY_STRATEGY
        _MEMORY_STRATEGY = strategy

    else:
        raise ValueError(
            "Please select one of the following options for a memory strategy "
            "'normal', 'moderate', 'aggressive'"
        )


def get_current_memory_strategy() -> str:
    """
    Gets the current memory_strategy

    Returns
    -------
    str
        Memory optimization strategy
    """
    return _MEMORY_STRATEGY


def _handle_xarray_memory(
    data_obj: Union[xr.Dataset, xr.DataArray], make_temp: bool = False
) -> Union[xr.Dataset, xr.DataArray]:
    """
    Executes memory strategy plan depending on the set memory strategy.

    Nothing happens in this method when the plan is normal, only if the strategy is moderate or aggressive does
    any memory optimization occur.  If the parameter make_temp is true a temporary file with a lzw compressed
    cloud optimized geotiff will be made, otherwise the original file will be loaded.

    Parameters
    ----------
    data_obj : Union[xr.Dataset, xr.DataArray]
        Xarray object to handle_memory
    make_temp: bool
        Store data in a temporary file if in memory

    Returns
    -------
    Union[xr.Dataset, xr.DataArray]
        Memory handled xarray object
    """

    # Check to see if it is already dask and what memory strategy is in place
    band_as_var = True if isinstance(data_obj, xr.Dataset) else False
    cache = True if _MEMORY_STRATEGY != "aggressive" else False

    if _MEMORY_STRATEGY == "normal" or _check_dask_array(data_obj):  # do nothing
        return data_obj
    elif (
        make_temp is False
    ):  # reload the file with cache on or off depending on strategy
        file_name = (
            data_obj["band_1"].encoding["source"]
            if band_as_var
            else data_obj.encoding["source"]
        )
        new_obj = _parse_string_attributes(
            rxr.open_rasterio(
                file_name,
                mask_and_scale=True,
                band_as_variable=band_as_var,
                cache=cache,
            )
        )
        del data_obj
        return new_obj

    else:  # make temporary file from lzw compressed COG and load with cache on or off depending on strategy
        dst_profile = cog_profiles.get("lzw")
        delete_file = not cache

        with NamedTemporaryFile(delete=True, suffix=".tif") as in_file:
            with NamedTemporaryFile(delete=delete_file, suffix=".tif") as out_file:
                data_obj.rio.to_raster(in_file.name, tiled=True, windowed=True)
                del data_obj
                cog_translate(in_file.name, out_file.name, dst_profile, in_memory=True)
                return _parse_string_attributes(
                    rxr.open_rasterio(
                        out_file.name,
                        mask_and_scale=True,
                        band_as_variable=band_as_var,
                        cache=cache,
                    )
                )


def _check_dask_array(original_map: Union[xr.DataArray, xr.Dataset]) -> bool:
    """
    Check whether map to be reprojected has dask data or not

    Parameters
    ----------
    original_map: Union[xr.DataArray, xr.Dataset]
        Map to be reprojected

    Returns
    -------
    bool
        Whether the data is a dask array
    """

    chunks = (
        original_map["band_1"].chunks
        if isinstance(original_map, xr.Dataset)
        else original_map.chunks
    )
    return chunks is not None


def _parse_string_attributes(
    obj: Union[xr.DataArray, xr.Dataset]
) -> Union[xr.DataArray, xr.Dataset]:
    """
    Parses string attributes stored in rasters

    Parameters
    ----------
    obj: Union[xr.DataArray, xr.Dataset]
        Xarray object with possible string attributes

    Returns
    -------
    Union[xr.DataArray, xr.Dataset]
        Object returned with parsed attributes
    """

    if "pairing_dictionary" in obj.attrs and isinstance(
        obj.attrs["pairing_dictionary"], str
    ):
        eval_str = ast.literal_eval(
            obj.attrs["pairing_dictionary"].replace("nan", '"nan"')
        )
        obj.attrs["pairing_dictionary"] = {
            (float(k[0]), float(k[1])): float(v) for k, v in eval_str.items()
        }

        if isinstance(obj, xr.Dataset):
            for var in obj.data_vars:
                obj[var].attrs["pairing_dictionary"] = obj.attrs["pairing_dictionary"]

    if isinstance(obj, xr.Dataset):
        for var in obj.data_vars:
            if "pairing_dictionary" in obj[var].attrs and isinstance(
                obj[var].attrs["pairing_dictionary"], str
            ):
                eval_str = ast.literal_eval(
                    obj[var].attrs["pairing_dictionary"].replace("nan", '"nan"')
                )
                obj[var].attrs["pairing_dictionary"] = {
                    (float(k[0]), float(k[1])): float(v) for k, v in eval_str.items()
                }

    return obj



"""
Test functionality for gval/homogenize/spatial_alignment.py
"""

# __all__ = ['*']
__author__ = "Fernando Aristizabal"

from pytest_cases import parametrize_with_cases
import xarray as xr

from gval.utils.loading_datasets import load_raster_as_xarray


@parametrize_with_cases("file_name")
def test_load_raster_as_xarray(file_name):
    """tests loading rasters as xarrays"""
    # NOTE: Do we need cases to test additional arguments if it is just wrapping rxr.open_rasterio()
    xr_map = load_raster_as_xarray(file_name)
    assert isinstance(
        xr_map, (xr.DataArray, xr.Dataset)
    ), "Did not load as xr.DataArray or xr.Dataset"
